#include "chil_global.h"

#define CLEAR_PIXEL(igate) \
  for (j=0;j<ray->ntotal;j++) *(ray->data[j]+igate) = chil_no_signal_value
#define CLEAR_FIELD_PIXEL(ifield, igate) \
  *(ray->data[ifield]+igate) = chil_no_signal_value


/* Subtract noise from the parameters Zh, Zdr and Ldr, given the noise
   at 1 km of the Zh, Zv and cross-polar channels.  These noise values
   can be determined by running the chilnoise program, which looks at
   data recorded above 14 km, which is assumed to be free from
   meteorological targets. */
int
CHILSubtractNoisePolar(Ray *ray, float nezh, float nezv, float nezhv,
		       float threshold)
{
  int i, j, iZh, iZdr, iLdr;
  float PZh, PZv, PZhv;
  float Pnezh, Pnezv, Pnezhv;

  char is_pci = 0;

  if (!ray->range_normalised || (iZh = CHILFindZH(ray)) < 0 ) { 
    return CHIL_NOTDONE;
  }

  iZdr = CHILFindField(ray, ZDR);
  iLdr = CHILFindField(ray, LDR);

  if (iZdr < 0 && iLdr < 0) {
    Pnezh = pow(10,0.1*nezh);
    for (i=0;i<ray->ngates;i++) {
      float *Zh = ray->data[iZh]+i;
      if (*Zh != chil_no_signal_value) {
	float range2=(ray->range[i]*ray->range[i]);
	PZh=pow(10,0.1 * (*Zh))/range2;
	if (PZh <= Pnezh*threshold) {
	  CLEAR_PIXEL(i);
	}
	else {
	  PZh -= Pnezh;
	  *Zh = 10.0 * log10(PZh*range2);
	}
      }
      else {
	/* if *Zh == chil_no_signal_value then the PCI system has
           already decided that there is no signal here */
	is_pci = 1;
      }
    }
  }
  else {
    Pnezh = pow(10,0.1*nezh);
    Pnezv = pow(10,0.1*nezv);
    Pnezhv = pow(10,0.1*nezhv);

    for (i=0;i<ray->ngates;i++) {
      float *Zh = ray->data[iZh]+i;
      if (*Zh != chil_no_signal_value) {
	float range2 = (ray->range[i]*ray->range[i]);
	PZh = pow(10,0.1* (*Zh))/range2;
	if (PZh <= Pnezh*threshold) {
	  CLEAR_PIXEL(i);
	}
	else {
	  PZh -= Pnezh;
	  if (iZdr >= 0) {
	    PZv = pow(10,0.1 * (*Zh-*(ray->data[iZdr]+i)))/range2;
	    if (PZv <= Pnezv*threshold) {
	      CLEAR_FIELD_PIXEL(iZdr, i);
	    }
	    else {
	      PZv -= Pnezv;
	      *(ray->data[iZdr]+i) = 10.0 * log10(PZh/PZv);
	    }
	  }
	  if (iLdr >= 0) {
	    PZhv = pow(10,0.1* (*Zh+*(ray->data[iLdr]+i)))/range2;
	    if (PZhv <= Pnezhv*threshold) {
	      CLEAR_FIELD_PIXEL(iLdr, i);
	    }
	    else {
	      PZhv -= Pnezhv;
	      *(ray->data[iLdr]+i) = 10.0 * log10(PZhv/PZh);
	    }
	  }
	  *Zh = 10.0 * log10(PZh*range2);
	}
      }
      else {
	/* if *Zh == chil_no_signal_value then the PCI system has
	   already decided that there is no signal here */
	is_pci = 1;
      }
    }
  }

  if (is_pci) {
    CHILPrintf("Warning: attempt to subtract noise from coherently-processed data\n");
  }

  return CHIL_OK;
}

/* Ground clutter tends to strongly depolarise the beam, so it can be
   removed by just throwing data away that has an Ldr greater than
   around -10 dB (i.e. a depolarisation of 10%). */
int
CHILSuppressClutterWithLdr(Ray *ray, float maxldr)
{
  int i, j, iLdr;
  if ((iLdr = CHILFindField(ray, LDR)) < 0) {
    return CHIL_NOTFOUND;
  }

  for (i=0;i<ray->ngates;i++) {
    if (*(ray->data[iLdr]+i) > maxldr && *(ray->data[iLdr]+i) 
	!= chil_no_signal_value) {
      CLEAR_PIXEL(i);
    }
  }
  return CHIL_OK;
}
