/* chil2nc.c -- converts Chilbolton-format radar data to netCDF
   A multitude of processing options are available

   Copyright (C) 2000 Robin Hogan
*/

#include <stdlib.h>
#include <stdio.h>
#include <netcdf.h>
#include <string.h>
#include <ctype.h>

#include "chil.h"
#include "nctools.h"
#include "version.h"

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    CHILPrintf("NetCDF error on line %d: %s\n", __LINE__, \
      nc_strerror(ncstatus)); \
    exit(-1); \
}

#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define MAX_STRING_LENGTH 512

static
void
add_attributes(int ncid, int fieldid, char *units, char *long_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
}

int
main(int argc, char **argv)
{
  int status, iarg = 0;
  ChilboltonData data;
  ChilboltonFile *chil_file = &(data.chil_file);
  RasterInfo *raster_info = &(data.raster_info);
  Ray *ray = &(data.ray);
  char aname[20];
  char *id = "";
  char *comment = NULL;
  char *user = NULL;
  float azimuth = 0, elevation = 90;
  int rayid, gateid, freqid, altid, latid, lonid;
  int ncid, rangeid, elevid, azimid, nezid, timeid, fieldid[MAX_FIELDS];
  int elevationid, azimuthid;
  size_t start[2], count[2];
  size_t tmp_index;
  int dimids[2];
  int index = 0;
  unsigned short tmp_uword;
  float tmp_float;
  unsigned short startgate, endgate, nfields = 1;
  char cloud_data, output_nez = 0, azel = 0, nopolarnames = 0;
  char fixed_elevation = 0, fixed_azimuth = 0;

  status = CHILInit(argc, argv, &data);

  if (argc < 2)
    {
      CHILUsage(&data);
      CHILPrintf("NetCDF-specific options\n"
		 "           -id s  optional string used in the filename\n"
		 "         -azel    output azim and elev vectors in cloud data\n"
		 "      -azimuth f  output azimuth as a scalar in cloud data\n"
		 "    -elevation f  output elevation as a scalar in cloud data\n"
		 " -nopolarnames    use Z instead of Zh for reflectivity\n"
		 "         -user s  name and email address of the user\n"
		 "      -comment s  comment to be added to each file\n");
      CHILOptions(&data);
      exit(0);
    }

  /* Look for netCDF-related arguments */
  
  while (++iarg < argc) {
    if (strcmp(argv[iarg],"-id") == 0) {
      if (argc > iarg+1) {
	id = argv[++iarg];
      }
      else {
	CHILPrintf("Error: id not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg],"-azel") == 0) {
      azel = 1;
    }
    else if (strcmp(argv[iarg],"-nopolarnames") == 0) {
      nopolarnames = 1;
    }
    else if (strcmp(argv[iarg],"-comment") == 0) {
      if (argc > iarg+1) {
	if (comment) {
	  CHILPrintf("Warning: comment will not appear: `%s'\n", comment);
	}
	comment = argv[++iarg];
      }
      else {
	CHILPrintf("Error: comment not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	if (user) {
	  CHILPrintf("Warning: user specified more than once: `%s' not used\n",
		     user);
	}
	user = argv[++iarg];
      }
      else {
	CHILPrintf("Error: user not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg], "-elevation") == 0) {
      if (argc > iarg+1) {
	elevation = atof(argv[++iarg]);
	fixed_elevation = 1;
      }
      else {
	CHILPrintf("Error: elevation not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg], "-azimuth") == 0) {
      if (argc > iarg+1) {
	azimuth = atof(argv[++iarg]);
	fixed_azimuth = 1;
      }
      else {
	CHILPrintf("Error: azimuth not specified\n");
	exit(1);
      }
    }
  }

  if (status == CHIL_OK)
    {

      while (1)
	{
	  int i, j;
	  char *version = CHIL_VERSION;
	  int failures = 0, maxfailures = 10;
	  char *tmp_name, *system_name;

	  while ((status = CHILFetchRay(&data)) != CHIL_OK)
	    {
	      if (status != CHIL_ENDOFBLOCK) {
		break;
	      }
	      if ((++failures) == maxfailures)
		{
		  CHILPrintf("Aborting: %d failed calls to CHILFetchRay"
			     " - empty scans?\n");
		  break;
		}
	    }
	  if (status != CHIL_OK)
	    {
	      break;
	    }

	  cloud_data = raster_info->cloud_data;

	  system_name = CHILSystemName(raster_info, NULL);

	  if (cloud_data)
	    {
#ifdef CHIL_USE_OLD_NAMES
	      sprintf(aname,"%04hu%02hu%02hu_%d%s.nc",
		      CHILTrueYear(ray->year),
		      ray->month, ray->day, (int) raster_info->frequency,
		      id);
#else
	      char *tmpstr = id;
	      char must_free = 0;

	      if (*id == '\0') {
		char *ch = strdup(system_name);
		tmpstr = ch;
		while (*ch) {
		  *ch = tolower(*ch);
		  ++ch;
		}
		must_free = 1;
	      }

	      sprintf(aname,"%04hu%02hu%02hu_%s.nc",
		      CHILTrueYear(ray->year),
		      ray->month, ray->day,
		      /*  (int) (raster_info->frequency+0.5), */
		      tmpstr);

	      if (must_free) {
		free(tmpstr);
	      }
#endif
	      if (data.process) {
		output_nez = 1;
	      }
	    }
	  else
	    {
	      sprintf(aname,"%04hur%03hus%02hu%s.nc", raster_info->file,
		      raster_info->raster, chil_file->iscan+1, id);
	      output_nez = 0;
	    }

	  /* Open netCDF file for writing */
	  handle_error(nc_create(aname, NC_CLOBBER, &ncid));
	  index = 0;
	  if (!data.quiet)
	    CHILPrintf("Writing %s\n", aname);

	  /* Add some global attributes */
	  tmp_name = CHILLocation(raster_info);
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "location",
				       strlen(tmp_name), tmp_name));
  
	  tmp_name = CHILSystemName(raster_info, NULL);
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "system",
				       strlen(tmp_name), tmp_name));
  
	  tmp_name = CHILLongSystemName(raster_info, NULL);
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "title",
				       strlen(tmp_name), tmp_name));
  
	  tmp_name = CHILScanType(raster_info, NULL);
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "scantype",
				       strlen(tmp_name), tmp_name));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "day", NC_SHORT,
					1, &(ray->day)));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "month", NC_SHORT,
					1, &(ray->month)));

	  tmp_uword = CHILTrueYear(ray->year);
	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "year", NC_SHORT,
					1, &tmp_uword));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "file",
					NC_SHORT,
					1, &(raster_info->file)));

	  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "raster",
					NC_SHORT,
					1, &(raster_info->raster)));
 	  if (!cloud_data) {
	    tmp_uword = chil_file->iscan + 1;
	    handle_error(nc_put_att_short(ncid, NC_GLOBAL, "scan",
					  NC_SHORT,
					  1, &tmp_uword));
	  }

	  /* Add a `history' attribute */
	  handle_error(nct_add_history(ncid, "Generated", user));

	  /* Add a `command_line' attribute */
	  handle_error(nct_add_command_line(ncid, argc, argv));

	  tmp_name = CHILGetOptions(&data);
	  /* The +1 is required in case strlen(tmp_name) == 0 */
	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "options",
				       strlen(tmp_name)+1, tmp_name));
	  CHILFree(tmp_name);

	  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "software_version",
				       strlen(version), version));

	  /* Add time/ray dimension */
	  handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, &rayid));

	  /* Add range/gate dimension */
	  endgate = ray->endgate;
	  nfields = MIN(ray->ntotal, data.maxfields);
	  startgate = ray->startgate;
	  handle_error(nc_def_dim(ncid, "range", endgate - startgate,
				  &gateid));

	  /* Add some scalar variables */
	  handle_error(nc_def_var(ncid, "frequency", NC_FLOAT, 0, dimids, 
				  &freqid));
	  add_attributes(ncid, freqid, "GHz", "Radar frequency");

	  handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, dimids, 
				  &latid));
	  add_attributes(ncid, latid, "degrees_north", "Latitude of antenna");

	  handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, dimids, 
				  &lonid));
	  add_attributes(ncid, lonid, "degrees_east", "Longitude of antenna");

	  handle_error(nc_def_var(ncid, "altitude", NC_FLOAT, 0, dimids, 
				  &altid));
	  add_attributes(ncid, altid, "m",
			 "Altitude of antenna above mean sea level");

	  if (!azel)
	    {
	      if (fixed_elevation) {
		handle_error(nc_def_var(ncid, "elevation", NC_FLOAT, 0, dimids,
					&elevationid));
		add_attributes(ncid, elevationid, "degrees",
			       "Elevation above horizon (fixed)");
	      }
	      if (fixed_azimuth) {
		handle_error(nc_def_var(ncid, "azimuth", NC_FLOAT, 0, dimids,
					&azimuthid));
		add_attributes(ncid, azimuthid, "degrees",
			       "Azimuth clockwise from due north (fixed)");
	      }
	    }

	  /* Add time coordinate variable */ 
	  dimids[0] = rayid;
	  handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, dimids,
				  &timeid));
	  add_attributes(ncid, timeid, "hours UTC",
			 "Decimal hours from midnight to the middle of each ray");

	  /* Add range coordinate variable */
	  dimids[0] = gateid;
	  handle_error(nc_def_var(ncid, "range", NC_FLOAT, 1,
				  dimids, &rangeid));
	  add_attributes(ncid, rangeid, "km",
			 "Range from antenna to the centre of each range gate");

	  /* If not cloud data, add elevation and azimuth */
	  dimids[0] = rayid;
	  if (!cloud_data || azel)
	    {
      	      handle_error(nc_def_var(ncid, "elev", NC_FLOAT, 1, dimids,
				    &elevid));
	      handle_error(nc_def_var(ncid, "azim", NC_FLOAT, 1, dimids,
				    &azimid));
	      if (data.averays > 1) {
		add_attributes(ncid, elevid, "degrees",
		     "Elevation of the centre of the ray above the horizon");
		add_attributes(ncid, azimid, "degrees",
		     "Azimuth of the centre of the ray clockwise from due north");
	      }
	      else {
		add_attributes(ncid, elevid, "degrees",
		     "Elevation of the end of the ray above the horizon");
		add_attributes(ncid, azimid, "degrees",
	             "Azimuth of the end of the ray clockwise from due north");
	      }
	    }

	  if (output_nez)
	    {
      	      handle_error(nc_def_var(ncid, "nez", NC_FLOAT, 1, dimids,
				    &nezid));
	      add_attributes(ncid, nezid, "dBZ",
			    "Noise-equivalent reflectivity factor at 1 km");
	    }
	  dimids[0] = rayid;
	  dimids[1] = gateid;

	  tmp_float = chil_no_signal_value;
	  for (i = 0; i < nfields; i++)
	    {
	      char *varcomments = NULL, *units_html = NULL;

	      tmp_name = CHILFieldName(ray, i, NULL);
	      if (nopolarnames && tmp_name[0] == 'Z'
		  && tmp_name[1] == 'h' && tmp_name[2] == '\0') {
		/* Call reflectivity Z not Zh */
		tmp_name = "Z";
	      }
	      handle_error(nc_def_var(ncid, tmp_name,
				      NC_FLOAT, 2, dimids, &fieldid[i]));

	      /* Add units and long name */
	      tmp_name = CHILFieldUnits(ray, i, NULL);
	      add_attributes(ncid, fieldid[i], tmp_name,
			     CHILFieldDescription(ray, i, NULL));
	      /* Add HTML units if different */
	      units_html = CHILFieldUnitsHTML(ray, i, NULL);
	      if (strcmp(tmp_name, units_html)) {
		handle_error(nc_put_att_text(ncid, fieldid[i], "units_html",
					     strlen(units_html), units_html));
	      }

	      handle_error(nc_put_att_float(ncid, fieldid[i],
					    "missing_value",
					    NC_FLOAT, 1, &tmp_float));
	      if (data.use_v1fold && ray->field_type[i] == V1) {
		handle_error(nc_put_att_float(ncid, fieldid[i],
					      "folding_velocity",
					      NC_FLOAT, 1, &data.v1fold));
	      }
	      else if (data.use_v2fold && ray->field_type[i] == V2) {
		handle_error(nc_put_att_float(ncid, fieldid[i],
					      "folding_velocity",
					      NC_FLOAT, 1, &data.v2fold));
	      }
	      else if (ray->field_type[i] == ZH) {
		handle_error(nc_put_att_float(ncid, fieldid[i],
					      "calibration_applied",
					      NC_FLOAT, 1, &data.Zcal));
	      }
	      else if (ray->field_type[i] == ZDR) {
		handle_error(nc_put_att_float(ncid, fieldid[i],
					      "calibration_applied",
					      NC_FLOAT, 1, &data.ZDRcal));
	      }
	      else if (ray->field_type[i] == LDR) {
		handle_error(nc_put_att_float(ncid, fieldid[i],
					      "calibration_applied",
					      NC_FLOAT, 1, &data.LDRcal));
	      }
	      else if (ray->field_type[i] == PHIDP) {
		handle_error(nc_put_att_float(ncid, fieldid[i],
					      "calibration_applied",
					      NC_FLOAT, 1, &data.PHIDPcal));
	      }

	      varcomments = CHILFieldInformation(&data, i);
	      if (varcomments) {
		handle_error(nc_put_att_text(ncid, fieldid[i], "comments",
					     strlen(varcomments), varcomments));
		CHILFree(varcomments);
	      }
	    }

	  if (comment) {
	    handle_error(nct_append_attribute(ncid, NC_GLOBAL, "comments",
					      comment, " "));
	  }

	  /* End define mode */
	  handle_error(nc_enddef(ncid));

	  /* Set the scalar variables */
	  tmp_index = 0;
	  handle_error(nc_put_var1_float(ncid, freqid, &tmp_index,
					 &(raster_info->frequency)));

	  tmp_float = CHILAltitude(raster_info);
	  handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &tmp_float));

	  tmp_float = CHILLatitude(raster_info);
	  handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &tmp_float));

	  tmp_float = CHILLongitude(raster_info);		       
	  handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &tmp_float));

	  if (fixed_elevation) {
	    handle_error(nc_put_var1_float(ncid, elevationid,
					   &tmp_index, &elevation));
	  }
	  if (fixed_azimuth) {
	    handle_error(nc_put_var1_float(ncid, azimuthid,
					   &tmp_index, &azimuth));
	  }

	  /* Set the range */
	  for (j = startgate; j < endgate; j++)
	    {
	      tmp_index = j - startgate;
	      tmp_float = ray->range[j] + 0.001 * ray->range_offset;
	      handle_error(nc_put_var1_float(ncid, rangeid,
					     &tmp_index, &tmp_float));
	    }

	  do
	    {
	      tmp_index = index;
	      handle_error(nc_put_var1_float(ncid, timeid, &tmp_index,
					   &(ray->utc)));
	      if (!cloud_data || azel)
		{
		  handle_error(nc_put_var1_float(ncid, elevid, &tmp_index,
					       &(ray->elevation)));
		  /* Make azimuth from due north rather than due west */
		  tmp_float = ray->azimuth - 90.0;
		  handle_error(nc_put_var1_float(ncid, azimid, &tmp_index,
					       &tmp_float));
		}

	      if (output_nez)
		{
		  handle_error(nc_put_var1_float(ncid, nezid, &tmp_index,
					       &(ray->nez)));
		}

	      for (i = 0; i < nfields; i++)
		{
		  start[0] = index;
		  start[1] = 0;
		  count[0] = 1;
		  count[1] = endgate - startgate;
		  handle_error(nc_put_vara_float(ncid, fieldid[i], start,
						 count, ray->data[i] 
						 + startgate));
		}

	      ++index;
	    }
	  while ((status = CHILFetchRay(&data)) == CHIL_OK);

	  handle_error(nc_close(ncid));

	  if (status == CHIL_ENDOFDATA)
	    {
	      break;
	    }
	}
      CHILEnd(&data);
    }
  else {
    CHILPrintf("Run with no options for more information\n");
    exit(1);
  }
  exit(0);
}
